import os
import pathlib
import tempfile
import shutil
from .reader import *

def _new_link_from_way(from_node,to_node, way):
    link = Link()
    link.osm_way_id = way.osm_way_id
    link.name = way.name
    link.link_type_name = way.railway
    link.from_node = from_node
    link.to_node = to_node
    link.railway = way.railway
    link.frequency = way.frequency
    link.gauge = way.gauge
    link.passenger_lines = way.passenger_lines
    link.railway_traffic_mode = way.railway_traffic_mode
    link.service = way.service
    link.electrified = way.electrified
    link.maxspeed = way.maxspeed
    link.maxspeed_designed = way.maxspeed_designed
    link.voltage = way.voltage
    link.usage = way.usage
    link.start_date = way.start_date
    link.railway_ctcs = way.railway_ctcs
    link.highspeed = way.highspeed
    link.from_node.outgoing_link_list.append(link)
    link.to_node.incoming_link_list.append(link)
    link.calculateLength()
    link.generate_geometry()
    return link

def _create_links(network, link_way_list):
    link_id = 0
    for way in link_way_list:
        for ref_node_id in range(len(way.ref_node_list) - 1):
            from_node = way.ref_node_list[ref_node_id]
            from_node.valid = True
            to_node = way.ref_node_list[ref_node_id + 1]
            to_node.valid = True
            link=_new_link_from_way(from_node,to_node, way)
            link.link_id=link_id
            network.link_dict[link_id]=link
            link_id+=1
    network.number_of_links = link_id

def _create_pois(network,POI_way_list):
    POI_list = []
    poi_id = 0
    for way in POI_way_list:
        poi = POI()
        poi.poi_id = poi_id
        poi.name = way.name
        poi.railway = way.railway
        poi.osm_way_id = way.osm_way_id
        ref_node_list = []
        for ref_node in way.ref_node_list:
            # ref_node.valid=True
            ref_node_list.append(ref_node.geometry)
        if way.ref_node_list[0].osm_node_id != way.ref_node_list[-1].osm_node_id:
            continue
        poi.ref_node_list = ref_node_list
        poi.generate_geometry()
        POI_list.append(poi)
        poi_id += 1
        network.POI_list.append(poi)

def _createNLPs(network,POIs):
    link_way_list = []
    POI_way_list = []
    for osm_way_id,way in network.osm_way_dict.items():
        if way.is_poi:
            POI_way_list.append(way)
        else:
            link_way_list.append(way)
    if len(link_way_list):
        _create_links(network, link_way_list)
    else:
        print('railway link not found')
    if POIs:
        if len(POI_way_list)>0:
            _create_pois(network,POI_way_list)
        else:
            print('railway POI not found')

def _filter_osm_node(network):
    node_dict = {}
    node_id = 0
    for osm_node_id, node in network.osm_node_dict.items():
        if node.valid and node.in_region:
            node.node_id = node_id
            node_dict[node_id] = node
            node_id += 1
    network.node_dict = node_dict

def _filter_osm_link(network):
    link_dict = {}
    link_id = 0
    for id, link in network.link_dict.items():
        from_node = link.from_node
        to_node = link.to_node
        if from_node.in_region and to_node.in_region:
            link_dict[link_id] = link
            link_id = link_id + 1
    network.link_dict = link_dict

def _filter_osm_poi(network):
    poi_list = []
    for poi in network.POI_list:
        if poi.geometry.within(network.bounds):
            poi_list.append(poi)
    network.POI_list = poi_list

def _buildNet(network,strict_mode=False,POIs=False):
    _createNLPs(network,POIs)
    _filter_osm_node(network)
    if strict_mode:
        _filter_osm_link(network)
        _filter_osm_poi(network)

def _check_filename(filename):
    path = pathlib.Path(filename)
    for ch in filename:
        if '\u4e00' <= ch <= '\u9fff':
            # 获取临时文件夹
            tmpdir = tempfile.gettempdir()
            if path.suffix == '.pbf':
                tmpfile=os.path.join(tmpdir,'map.osm.pbf')
            else:
                tmpfile=os.path.join(tmpdir,'map.osm')
            shutil.copy(filename,tmpfile)
            return tmpfile,True
    return filename,False

def get_network_from_file(filename='map.osm',bbox=None,strict_mode=False,POIs=False):
    filename,is_tmp=_check_filename(filename)
    network=readOSMFile(filename=filename,strict_mode=strict_mode,bbox=bbox)
    _buildNet(network=network,strict_mode=strict_mode,POIs=POIs)
    if is_tmp:
        os.remove(filename)
    return network


import osmium
from .classes import *
from .settings import *


def _getBounds(filename, bbox):
    try:
        f = osmium.io.Reader(filename)
    except:
        sys.exit(f'Error: filepath {filename} contains invalid characters. Please use english characters only')

    if bbox is None:
        header = f.header()
        box = header.box()
        bottom_left = box.bottom_left
        top_right = box.top_right
        try:
            minlat, minlon = bottom_left.lat, bottom_left.lon
            maxlat, maxlon = top_right.lat, top_right.lon
        except:
            minlat, minlon, maxlat, maxlon = default_bounds['minlat'], default_bounds['minlon'], default_bounds['maxlat'], default_bounds['maxlon']
    else:
        minlat, maxlat,minlon, maxlon = bbox

    bounds = geometry.Polygon([(minlon, maxlat), (maxlon, maxlat), (maxlon, minlat), (minlon, minlat)])
    return bounds

class NWRHandler(osmium.SimpleHandler):
    def __init__(self):
        osmium.SimpleHandler.__init__(self)

        self.strict_mode = True
        self.bounds = None

        self.osm_node_dict = {}
        self.osm_node_id_list = []
        self.osm_node_coord_list = []

        self.osm_way_dict = {}

    def node(self, n):
        node = Node()
        node.osm_node_id = str(n.id)
        lon, lat = n.location.lon, n.location.lat
        node.geometry = geometry.Point((round(lon, lonlat_precision), round(lat, lonlat_precision)))
        node.x_coord=lon
        node.y_coord=lat
        if self.strict_mode:
            if not node.geometry.within(self.bounds):
                node.in_region = False

        self.osm_node_id_list.append(node.osm_node_id)
        self.osm_node_coord_list.append((lon, lat))

        node.osm_railway = n.tags.get('railway')
        node.access=n.tags.get('access')
        node.description=n.tags.get('description')
        node.level_crossing=n.tags.get('level_crossing')
        self.osm_node_dict[node.osm_node_id] = node
        del n

    def way(self, w):
        way = Way()
        way.osm_way_id = str(w.id)
        way.ref_node_id_list = [str(node.ref) for node in w.nodes]
        way.railway = w.tags.get('railway')
        landuse=w.tags.get('landuse')
        if way.railway =='rail':
            way.name = w.tags.get('name')
            way.electrified = w.tags.get('electrified')
            way.frequency = w.tags.get('frequency')
            way.highspeed = w.tags.get('highspeed')
            way.maxspeed = w.tags.get('maxspeed')
            way.maxspeed_designed = w.tags.get('maxspeed:designed')
            way.passenger_lines = w.tags.get('passenger_lines')
            way.railway_ctcs = w.tags.get('railway:ctcs')
            way.railway_traffic_mode = w.tags.get('railway:traffic_mode')
            way.start_date = w.tags.get('start_date')
            way.usage = w.tags.get('usage')
            way.voltage = w.tags.get('voltage')
            way.service = w.tags.get('service')
            way.gauge = w.tags.get('gauge')
            self.osm_way_dict[way.osm_way_id] = way
        elif way.railway =='platform':
                way.is_poi=True
                way.name = w.tags.get('name')
                self.osm_way_dict[way.osm_way_id] = way
        elif way.railway is None and landuse is not None:
            if landuse in railway_poi_set:
                way.landuse=landuse
                if w.tags.get('name'):
                    way.name=w.tags.get('name')
                elif w.tags.get('operator'):
                    way.name=w.tags.get('operator')
                if way.name:
                    way.is_poi=True
                    self.osm_way_dict[way.osm_way_id] = way
            elif landuse == 'industrial' and w.tags.get('industrial'):
                industrial=w.tags.get('industrial')
                if industrial in railway_poi_set:
                    if w.tags.get('name'):
                        way.name=w.tags.get('name')
                    elif w.tags.get('operator'):
                        way.name = w.tags.get('operator')
                    if way.name:
                        way.is_poi = True
                        way.landuse='railway'
                        self.osm_way_dict[way.osm_way_id] = way
        del w

def _processWays(net, h):
    for osm_way_id, osm_way in h.osm_way_dict.items():
        try:
            osm_way.ref_node_list = [net.osm_node_dict[ref_node_id] for ref_node_id in osm_way.ref_node_id_list]
            net.osm_way_dict[osm_way_id] = osm_way
        except KeyError as e:
            print(f'  warning: ref node {e} in way {osm_way_id} is not defined, way {osm_way_id} will not be imported')

def readOSMFile(filename, strict_mode, bbox):
    net = Network()

    net.bounds = _getBounds(filename, bbox)

    h = NWRHandler()
    h.strict_mode = strict_mode
    h.bounds = net.bounds
    h.apply_file(filename)

    net.osm_node_dict=h.osm_node_dict
    _processWays(net,h)


    return net


from .utils import get_distance_from_coord
from shapely import geometry

class Node:
    def __init__(self):
        self.name = None
        self.node_id = 0
        self.osm_node_id = None

        self.x_coord=None
        self.y_coord=None
        self.geometry = None

        self.level_crossing=None
        self.access=None
        self.description=None
        self.osm_railway=None

        self.in_region=True
        self.valid = False

        self.incoming_link_list = []
        self.outgoing_link_list = []

class Link:
    def __init__(self):
        self.link_id = 0
        self.osm_way_id = None      # str
        self.name = ''
        self.loc_name=''
        self.link_type_name = None
        self.from_node = None
        self.to_node = None

        self.railway=None
        self.geometry_str=None
        self.geometry = None

        self.length=None
        self.electrified = None
        self.frequency = None
        self.highspeed = None
        self.maxspeed = None
        self.maxspeed_designed = None
        self.passenger_lines = None
        self.railway_ctcs = None
        self.railway_traffic_mode = None
        self.start_date = None
        self.usage = None
        self.voltage = None
        self.gauge=None
        self.service=None

    def calculateLength(self):
        self.length=get_distance_from_coord(self.from_node.x_coord,self.from_node.y_coord,
                                              self.to_node.x_coord,self.to_node.y_coord) * 1000
    def generate_geometry(self):
        self.geometry=geometry.LineString([self.from_node.geometry,self.to_node.geometry])

class POI:
    def __init__(self):
        self.poi_id = 0
        self.osm_way_id = None
        self.name = None
        self.geometry = None
        self.ref_node_list=[]

    def generate_geometry(self):
        self.geometry = geometry.Polygon(self.ref_node_list)


class Way:
    def __init__(self):
        self.osm_way_id = None          # string
        self.name=''
        self.railway = ''

        self.electrified=''
        self.frequency=None
        self.highspeed=None
        self.maxspeed=None
        self.maxspeed_designed=None
        self.passenger_lines=None
        self.railway_ctcs=None
        self.railway_traffic_mode=None
        self.start_date=None
        self.usage=None
        self.voltage=None
        self.service=None
        self.landuse=None
        self.gauge=None
        self.is_poi =None
        self.layer=None
        self.ref_node_list = []

class Network:
    def __init__(self):
        self.bounds = None
        self.central_lon = 0.0
        self.central_lat = 0.0

        self.osm_node_dict = {}
        self.osm_way_dict = {}

        self.node_dict = {}
        self.link_dict = {}
        self.POI_list = []

        self.central_lon=None

        self.number_of_links=0


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection,PolyCollection

def showNetwork(network,savefig=None):
    node_x_coords=[]
    node_y_coords=[]
    link_coords=[]
    poi_coords=[]

    for _,node in network.node_dict.items():
        node_x_coords.append(node.x_coord)
        node_y_coords.append(node.y_coord)

    for _,link in network.link_dict.items():
        coords = list(link.geometry.coords)
        link_coords.append(np.array(coords))

    if len(network.POI_list):
        for poi in network.POI_list:
            coords = list(poi.geometry.exterior.coords)
            poi_coords.append(np.array(coords))

    fig, ax = plt.subplots(figsize=(12, 8))
    # plot network nodes
    ax.scatter(node_x_coords, node_y_coords, marker='o', c='red', s=10, zorder=1)
    # plot network links
    ax.add_collection(LineCollection(link_coords, colors='orange', linewidths=1, zorder=2))
    # plot network pois
    if len(poi_coords):
        coll = PolyCollection(poi_coords, alpha=0.7, zorder=0)
        ax.add_collection(coll)
    # set axis
    ax.autoscale_view()
    plt.xlabel('x_coord')
    plt.ylabel('y_coord')
    plt.tight_layout()
    # show fig
    plt.show()
    # save fig
    if savefig:
        try:
            figname = savefig['filename'] if 'filename' in savefig.keys() else 'network.png'
            dpi = savefig['dpi'] if 'dpi' in savefig else 300
            fig.savefig(figname, dpi=dpi, bbox_inches='tight')
        except Exception as e:
            print(e)


import math
import os
import pkg_resources
import re
import copy
import pathlib
import string
import fuzzywuzzy.fuzz
import difflib
import pickle
import collections
import numpy as np

def update_nested_dict(source_dict, updates):
    for key, val in updates.items():
        if isinstance(val, collections.abc.Mapping) or isinstance(val, dict):
            source_dict[key] = update_nested_dict(source_dict.get(key, {}), val)

        elif isinstance(val, list):
            source_dict[key] = (source_dict.get(key, []) + val)

        else:
            source_dict[key] = updates[key]

    return source_dict

def remove_punctuation(raw_txt, rm_whitespace=False):

    try:
        txt = raw_txt.translate(str.maketrans('', '', string.punctuation))

    except Exception as e:
        print(e)
        txt = ''.join(x for x in raw_txt if x not in string.punctuation)

    if rm_whitespace:
        txt = ' '.join(txt.split())

    return txt

def find_similar_str(str_x, lookup_list, n=1, processor='fuzzywuzzy', ignore_punctuation=True,
                     **kwargs):
    assert processor in ('difflib', 'fuzzywuzzy'), \
        "Options for `processor` include \"difflib\" and \"fuzzywuzzy\"."

    n_ = len(lookup_list) if n is None else n

    if processor == 'fuzzywuzzy':

        l_distances = [fuzzywuzzy.fuzz.token_set_ratio(str_x, a, **kwargs) for a in lookup_list]

        if sum(l_distances) == 0:
            sim_str = None
        else:
            if n_ == 1:
                sim_str = lookup_list[l_distances.index(max(l_distances))]
            else:
                sim_str = [lookup_list[i] for i in np.argsort(l_distances)[::-1]][:n_]

    elif processor == 'difflib':
        if not ignore_punctuation:
            str_x_ = str_x.lower()
            lookup_list_ = {x_.lower(): x_ for x_ in lookup_list}
        else:
            str_x_ = remove_punctuation(str_x.lower())
            lookup_list_ = {remove_punctuation(x_.lower()): x_ for x_ in lookup_list}

        sim_str_ = difflib.get_close_matches(str_x_, lookup_list_.keys(), n=n_, **kwargs)

        if not sim_str_:
            sim_str = None
        else:
            sim_str = lookup_list_[sim_str_[0]]

    else:
        sim_str = None

    return sim_str

def _get_specific_filepath_info(path_to_file, verbose=False, verbose_end=" ... ", ret_info=False):

    abs_path_to_file = pathlib.Path(path_to_file).absolute()
    assert not abs_path_to_file.is_dir(), "The input for `path_to_file` may not be a file path."

    filename = pathlib.Path(abs_path_to_file).name if abs_path_to_file.suffix else ""

    try:
        rel_path = pathlib.Path(os.path.relpath(abs_path_to_file.parent))

        if rel_path == rel_path.parent:
            rel_path = abs_path_to_file.parent
            msg_fmt = "{} \"{}\""
        else:
            msg_fmt = "{} \"{}\" {} \"{}\\\""
            # The specified path exists?
            os.makedirs(abs_path_to_file.parent, exist_ok=True)

    except ValueError:
        if verbose == 2:
            warn_msg = "Warning: \"{}\" is outside the current working directory".format(
                str(abs_path_to_file.parent))
            print(warn_msg)

        rel_path = abs_path_to_file.parent
        msg_fmt = "{} \"{}\""

    if verbose:
        if os.path.isfile(abs_path_to_file):
            status_msg, prep = "Updating", "at"
        else:
            status_msg, prep = "Saving", "to"

        if verbose_end:
            verbose_end_ = verbose_end
        else:
            verbose_end_ = "\n"

        if rel_path == rel_path.parent:
            print(msg_fmt.format(status_msg, filename), end=verbose_end_)
        else:
            print(msg_fmt.format(status_msg, filename, prep, rel_path), end=verbose_end_)

    if ret_info:
        return rel_path, filename

def save_pickle(pickle_data, path_to_pickle, mode='wb', verbose=False, **kwargs):

    _get_specific_filepath_info(path_to_pickle, verbose=verbose, ret_info=False)

    try:
        pickle_out = open(path_to_pickle, mode=mode)
        pickle.dump(pickle_data, pickle_out, **kwargs)
        pickle_out.close()
        print("Done.") if verbose else ""

    except Exception as e:
        print("Failed. {}.".format(e))

def load_pickle(path_to_pickle, mode='rb', verbose=False, **kwargs):

    if verbose:
        print("Loading \"{}\"".format(os.path.relpath(path_to_pickle)), end=" ... ")

    try:
        pickle_in = open(path_to_pickle, mode=mode)
        pickle_data = pickle.load(pickle_in, **kwargs)
        pickle_in.close()
        print("Done.") if verbose else ""
        return pickle_data

    except Exception as e:
        print("Failed. {}".format(e)) if verbose else ""

def confirmed(prompt=None, confirmation_required=True, resp=False):

    if confirmation_required:
        if prompt is None:
            prompt_ = "Confirmed? "
        else:
            prompt_ = copy.copy(prompt)

        if resp is True:  # meaning that default response is True
            prompt_ = "{} [{}]|{}: ".format(prompt_, "Yes", "No")
        else:
            prompt_ = "{} [{}]|{}: ".format(prompt_, "No", "Yes")

        ans = input(prompt_)
        if not ans:
            return resp

        if re.match('[Yy](es)?', ans):
            return True
        if re.match('[Nn](o)?', ans):
            return False

    else:
        return True

def validate_input_data_dir(input_data_dir=None, msg="Invalid input!", sub_dir=""):
    if input_data_dir:
        assert isinstance(input_data_dir, str), msg

        if not os.path.isabs(input_data_dir):  # Use default file directory
            data_dir_ = cd(input_data_dir.strip('.\\.'))

        else:
            data_dir_ = os.path.realpath(input_data_dir.lstrip('.\\.'))
            assert os.path.isabs(input_data_dir), msg

    else:
        data_dir_ = cd(sub_dir) if sub_dir else cd()

    return data_dir_

def cd(*sub_dir, mkdir=False, cwd=None, back_check=False, **kwargs):
    # Current working directory
    path = os.getcwd() if cwd is None else copy.copy(cwd)

    if back_check:
        while not os.path.exists(path):
            path = os.path.dirname(path)

    for x in sub_dir:
        path = os.path.dirname(path) if x == ".." else os.path.join(path, x)

    if mkdir:
        path_to_file, ext = os.path.splitext(path)

        if ext == '':
            os.makedirs(path_to_file, exist_ok=True, **kwargs)
        else:
            os.makedirs(os.path.dirname(path_to_file), exist_ok=True, **kwargs)

    return path

def cd_dat(*sub_dir, dat_dir="dat", mkdir=False, **kwargs):
    path = pkg_resources.resource_filename(__name__, dat_dir)

    for x in sub_dir:
        path = os.path.join(path, x)

    if mkdir:
        path_to_file, ext = os.path.splitext(path)

        if ext == '':
            os.makedirs(path_to_file, exist_ok=True, **kwargs)
        else:
            os.makedirs(os.path.dirname(path_to_file), exist_ok=True, **kwargs)

    return path

def cd_dat_geofabrik(*sub_dir, mkdir=False, **kwargs):

    path = cd("dat_Geofabrik", *sub_dir, mkdir=mkdir, **kwargs)

    return path

def cd_dat_bbbike(*sub_dir, mkdir=False, **kwargs):

    path = cd("dat_BBBike", *sub_dir, mkdir=mkdir, **kwargs)

    return path

def get_distance_from_coord(lon1, lat1, lon2, lat2):
    # return km
    lon1, lat1, lon2, lat2 = map(math.radians, [lon1, lat1, lon2, lat2])
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
    c = 2 * math.asin(math.sqrt(a))
    r = 6371
    return c * r

def validate_download_dir(dirname):
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    return dirname


import random
import requests
import sys
import time

def overpass_homepage():
    return 'https://overpass-api.de/api/map?bbox='
def geofabrik_homepage():
    return 'http://download.geofabrik.de/'
def bbbike_homepage():
    return 'http://download.bbbike.org/osm/bbbike/'

def fake_requests_headers():
    UA = [
        "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.153 Safari/537.36",
        "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:30.0) Gecko/20100101 Firefox/30.0",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) AppleWebKit/537.75.14 (KHTML, like Gecko) Version/7.0.3 Safari/537.75.14",
        "Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; Win64; x64; Trident/6.0)",
        'Mozilla/5.0 (Windows; U; Windows NT 5.1; it; rv:1.8.1.11) Gecko/20071127 Firefox/2.0.0.11',
        'Opera/9.25 (Windows NT 5.1; U; en)',
        'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)',
        'Mozilla/5.0 (compatible; Konqueror/3.5; Linux) KHTML/3.5.5 (like Gecko) (Kubuntu)',
        'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.0.12) Gecko/20070731 Ubuntu/dapper-security Firefox/1.5.0.12',
        'Lynx/2.8.5rel.1 libwww-FM/2.14 SSL-MM/1.4.1 GNUTLS/1.2.9',
        "Mozilla/5.0 (X11; Linux i686) AppleWebKit/535.7 (KHTML, like Gecko) Ubuntu/11.04 Chromium/16.0.912.77 Chrome/16.0.912.77 Safari/535.7",
        "Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:10.0) Gecko/20100101 Firefox/10.0 "
    ]
    return random.choice(UA)

"""determine the coordinate range of target area"""
def get_subregion_download_range(area):
    html = requests.get(f'https://nominatim.openstreetmap.org/search?format=json&q={area}')
    if html.text=='[]':
        print('Unable to query the address,please try other keys')
        sys.exit(0)
    else:
        info = eval(html.text)[0]
        return info['boundingbox']

"""download .osm file for OverpassDownloader"""
def download_osmfile_from_url(url, osm_filename):
    user_agent = fake_requests_headers()
    # Streaming, so we can iterate over the response
    resp = requests.get(url, stream=True, headers={"User-Agent": user_agent})
    if resp.status_code == 429:
        print("\033[31m Error! Too many requests, please try again later, or try to use random header.[0m", end=" ")
        return
    if resp.status_code==400:
        print(resp.text)
        return
    total_size = resp.content.__sizeof__()-33 # Total size in bytes
    block_size = 1024*1024
    wrote = 0

    with open(osm_filename, mode='wb') as f:
        for data in resp.iter_content(block_size, decode_unicode=True):
            wrote = wrote + len(data)
            print("\033[31m \r download progress：%.2f MB(%.2f MB)\033[0m" % (wrote /block_size,total_size/block_size), end=" ")
            try:
                f.write(data)
            except TypeError:
                f.write(data.encode())
            time.sleep(0.2)
        f.close()
    resp.close()
    print()
    if total_size != 0 and wrote != total_size:
        print("ERROR, something went wrong!")




import csv
import os
from .utils import validate_download_dir

def saveNetwork(network, output_folder='csvfile', enconding=None):
    validate_download_dir(output_folder)
    if output_folder:
        validate_download_dir(output_folder)
        node_filepath = os.path.join(output_folder, 'node.csv')
        link_filepath = os.path.join(output_folder, 'link.csv')
        poi_filepath = os.path.join(output_folder, 'poi.csv')
    else:
        node_filepath = 'node.csv'
        link_filepath = 'link.csv'
        poi_filepath = 'poi.csv'

    try:
        if enconding is None:
            outfile = open(node_filepath, 'w', newline='', errors='ignore')
        else:
            outfile = open(node_filepath, 'w', newline='', errors='ignore', encoding=enconding)

        write = csv.writer(outfile)
        write.writerow(['name', 'node_id', 'osm_node_id', 'osm_railway', 'level_crossing', 'access', 'description',
                        'x_coord', 'y_coord', 'geometry'])

        for node_id, node in network.node_dict.items():
            name = node.name if node.name else ''
            level_crossing = node.level_crossing if node.level_crossing else ''
            access = node.access if node.access else ''
            description = node.description if node.description else ''
            railway = node.osm_railway if node.osm_railway else ''
            geometry = node.geometry.wkt
            line = [name, node.node_id, node.osm_node_id, railway, level_crossing, access, description,
                    node.x_coord, node.y_coord, geometry]
            write.writerow(line)
        outfile.close()
    except PermissionError:
        print('node.csv may be locked by other programs. please release it then try again')

    try:
        if enconding is None:
            outfile = open(link_filepath, 'w', newline='', errors='ignore')
        else:
            outfile = open(link_filepath, 'w', newline='', errors='ignore', encoding=enconding)

        write = csv.writer(outfile)
        write.writerow(['name', 'link_id', 'osm_way_id', 'from_node_id', 'to_node_id', 'link_type_name', 'electrified',
                        'railway','frequency', 'highspeed', 'maxspeed','maxspeed_designed', 'passenger_lines', 'railway_ctcs',
                        'railway_traffic_mode', 'start_date','usage', 'voltage', 'gauge', 'service', 'length', 'geometry'])
        for link_id, link in network.link_dict.items():
            name = link.name if link.name else ''
            link_type_name = link.link_type_name if link.link_type_name else ''
            electrified = link.electrified if link.electrified else ''
            railway=link.railway if link.railway else ''
            frequency = link.frequency if link.frequency else ''
            highspeed = link.highspeed if link.highspeed else ''
            maxspeed = link.maxspeed if link.maxspeed else ''
            maxspeed_designed = link.maxspeed_designed if link.maxspeed_designed else ''
            passenger_lines = link.passenger_lines if link.passenger_lines else ''
            railway_ctcs = link.railway_ctcs if link.railway_ctcs else ''
            railway_traffic_mode = link.railway_traffic_mode if link.railway_traffic_mode else ''
            start_date = link.start_date if link.start_date else ''
            usage = link.usage if link.usage else ''
            voltage = link.voltage if link.voltage else ''
            gauge = link.gauge if link.gauge else ''
            service = link.service if link.service else ''
            line = [name, link.link_id, link.osm_way_id, link.from_node.node_id, link.to_node.node_id,
                    link_type_name, electrified, railway,frequency,highspeed, maxspeed, maxspeed_designed, passenger_lines,
                    railway_ctcs, railway_traffic_mode,start_date, usage, voltage, gauge, service, link.length,
                    link.geometry.wkt]
            write.writerow(line)
        outfile.close()
    except PermissionError:
        print('link.csv may be locked by other programs. please release it then try again')

    try:
        if len(network.POI_list):
            if enconding is None:
                outfile = open(poi_filepath, 'w', newline='', errors='ignore')
            else:
                outfile = open(poi_filepath, 'w', newline='', errors='ignore', encoding=enconding)

            write = csv.writer(outfile)
            write.writerow(['name', 'poi_id', 'osm_way_id', 'railway','geometry'])

            for poi in network.POI_list:
                name = ' ' + poi.name if poi.name else ''
                line = [name, poi.poi_id, poi.osm_way_id, poi.railway, poi.geometry.wkt]
                write.writerow(line)
            outfile.close()
    except PermissionError:
        print('poi.csv may be locked by other programs. please release it then try again')




from .download_by_Overpass import download_osm_data_from_overpass
from .download_by_BBBike import download_osm_data_from_bbbike
from .download_by_BBBike import download_subregion_data_from_bbbike
from .download_by_BBBike import update_bbbike_database
from .download_by_Geofabrik import download_osm_data_from_geofabrik
from .download_by_Geofabrik import download_subregion_data_from_geofabrik
from .download_by_Geofabrik import update_geofabrik_database
from .network import get_network_from_file
from .writer import saveNetwork
from .plotter import showNetwork

__all__=[
    'download_osm_data_from_overpass',
    'download_osm_data_from_geofabrik',
    'download_subregion_data_from_geofabrik',
    'update_geofabrik_database',
    'download_osm_data_from_bbbike',
    'download_subregion_data_from_bbbike',
    'update_bbbike_database',
    'get_network_from_file',
    'showNetwork',
    'saveNetwork'
]

__version__ = '0.0.3'
__author__ = u'Jiawei Lu, Qian Fu,Zanyang Cui, Junhua Chen'
__email__ = 'jiaweil9@asu.edu, qian.fu@outlook.com, zanyangcui@outlook.com, cjh@bjtu.edu.cn'


import urllib
from .web import *

"""setting for Geofabrik downloader (Geofabrik OpenStreetMap data extracts)"""
GeofabrikURL = geofabrik_homepage()
GeofabrikDownloadIndexURL = urllib.parse.urljoin(GeofabrikURL, 'index-v1.json')
GeofabrikValidFileFormats = [".osm.pbf", ".shp.zip", ".osm.bz2"]
GeofabrikDownloadIndexName = 'Geofabrik index of all downloads'
GeofabrikContinentSubregionTableName = 'Geofabrik continent subregions'
GeofabrikRegionSubregionTier = 'Geofabrik region-subregion tier'
GeofabrikDownloadsCatalogue = 'Geofabrik downloads catalogue'
GeofabrikSubregionNameList = 'Geofabrik subregion name list'

"""setting for BBBike downloader (BBBike OpenStreetMap data extracts)"""
BBBikeURL = bbbike_homepage()
BBBikeURLCities = 'https://raw.githubusercontent.com/wosch/bbbike-world/world/etc/cities.txt'
BBBikeCitiesNames = 'BBBike cities'
BBBikeURLCitiesCoordinates = 'https://raw.githubusercontent.com/wosch/bbbike-world/world/etc/cities.csv'
BBBikeCitiesCoordinates = 'BBBike cities coordinates'
BBBikeSubregionCatalogue = 'BBBike subregion catalogue'
BBBikeSubregionNameList = 'BBBike subregion name list'
BBBikeDownloadDictName = 'BBBike download dictionary'


lonlat_precision = 7
default_bounds = {'minlat':-90.0, 'minlon':-180.0, 'maxlat':90.0, 'maxlon':180.0}
railway_poi_set = ['depot','station','workshop','halt','interlocking','junction','spur_junction','terminal','platform']


import io
import bs4
import csv
import pandas as pd
from .utils import *
from .settings import *

def _get_list_of_cities(update=False, confirmation_required=True, verbose=False):
    path_to_pickle = cd_dat(BBBikeCitiesNames.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_pickle) and not update:
        cities_names = load_pickle(path_to_pickle)

    else:
        if confirmed("To collect {}?".format(BBBikeCitiesNames), confirmation_required=confirmation_required):

            try:
                cities_names_ = pd.read_csv(BBBikeURLCities, header=None)
                cities_names = list(cities_names_.values.flatten())

                save_pickle(cities_names, path_to_pickle, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                cities_names = None

        else:
            if verbose:
                print("No data of \"{}\" is available.".format(BBBikeCitiesNames))
            cities_names = None

    return cities_names

def _get_coordinates_of_cities(update=False, confirmation_required=True, verbose=False):

    path_to_pickle = cd_dat(BBBikeCitiesCoordinates.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_pickle) and not update:
        cities_coordinates = load_pickle(path_to_pickle)

    else:
        if confirmed("To collect {}?".format(BBBikeCitiesCoordinates),
                     confirmation_required=confirmation_required):

            try:
                csv_temp = urllib.request.urlopen(BBBikeURLCitiesCoordinates)
                csv_file = list(csv.reader(io.StringIO(csv_temp.read().decode('utf-8')),delimiter=':'))

                csv_data = [[x.strip().strip('\u200e').replace('#', '') for x in row] for row in csv_file[5:-1]]
                column_names = [x.replace('#', '').strip().capitalize() for x in csv_file[0]]
                cities_coords = pd.DataFrame(csv_data, columns=column_names)

                coordinates = cities_coords.Coord.str.split(' ').apply(pd.Series)
                coords_cols = ['ll_longitude', 'll_latitude1','ur_longitude', 'ur_latitude']
                coordinates.columns = coords_cols

                cities_coords.drop(['Coord'], axis=1, inplace=True)

                cities_coordinates = pd.concat([cities_coords, coordinates], axis=1)

                cities_coordinates.dropna(subset=coords_cols, inplace=True)

                save_pickle(cities_coordinates, path_to_pickle, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                cities_coordinates = None

        else:
            if verbose:
                print("No data of \"{}\" is available.".format(BBBikeCitiesCoordinates))
            cities_coordinates = None

    return cities_coordinates

def _get_subregion_catalogue(update=False, confirmation_required=True,verbose=False):

    path_to_pickle = cd_dat(BBBikeSubregionCatalogue.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_pickle) and not update:
        subregion_catalogue = load_pickle(path_to_pickle)

    else:
        if confirmed("To collect {}?".format(BBBikeSubregionCatalogue),confirmation_required=confirmation_required):

            try:
                bbbike_subregion_catalogue_ = pd.read_html(BBBikeURL, header=0, parse_dates=['Last Modified'])
                subregion_catalogue = bbbike_subregion_catalogue_[0].drop(0).drop(['Size', 'Type'], axis=1)
                subregion_catalogue.Name = subregion_catalogue.Name.map(lambda x: x.strip('/'))
                user_agent = fake_requests_headers()
                source = requests.get(BBBikeURL, headers={"User-Agent": user_agent})
                table_soup = bs4.BeautifulSoup(source.text, 'lxml').find('table')
                urls = [urllib.parse.urljoin(BBBikeURL, x.get('href')) for x in table_soup.find_all('a')[1:]]

                subregion_catalogue['URL'] = urls

                save_pickle(subregion_catalogue, path_to_pickle, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                subregion_catalogue = None

        else:
            if verbose:
                print("No data of \"{}\" is available.".format(BBBikeSubregionCatalogue))
            subregion_catalogue = None

    return subregion_catalogue

def _get_list_of_subregion_names(update=False, confirmation_required=True,verbose=False):

    path_to_name_list = cd_dat(BBBikeSubregionNameList.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_name_list) and not update:
        subregion_name_list = load_pickle(path_to_name_list)

    else:
        if confirmed("To get {}?".format(BBBikeSubregionNameList),confirmation_required=confirmation_required):

            subregion_catalogue = _get_subregion_catalogue(update, confirmation_required=False, verbose=verbose)

            subregion_name_list = subregion_catalogue.Name.to_list()

            save_pickle(subregion_name_list, path_to_name_list, verbose=verbose)

        else:
            subregion_name_list = []
            if verbose:
                print("No data of {} is available.".format(BBBikeSubregionNameList))

    return subregion_name_list

def _validate_input_subregion_name(subregion_name):

    assert isinstance(subregion_name, str)

    bbbike_subregion_names =_get_list_of_subregion_names()

    subregion_name_ = find_similar_str(subregion_name, bbbike_subregion_names)

    return subregion_name_

def _get_subregion_download_catalogue(subregion_name, confirmation_required=True,verbose=False):
    subregion_name_ = _validate_input_subregion_name(subregion_name)

    if confirmed("Confirmed to collect the download catalogue for {}?".format(subregion_name_),
                 confirmation_required=confirmation_required):

        try:
            if confirmation_required:
                print("In progress", end=" ... ") if verbose else ""
            else:
                print(f" {subregion_name_}", end=" ... ") if verbose else ""

            url = urllib.parse.urljoin(BBBikeURL, subregion_name_ + '/')
            user_agent = fake_requests_headers()
            source = requests.get(url, headers={"User-Agent": user_agent})

            source_soup = bs4.BeautifulSoup(source.text, 'lxml')
            download_links_class = source_soup.find_all(name='a', attrs={'class': ['download_link', 'small']})

            def parse_dlc(dlc):
                dlc_href = dlc.get('href')  # URL
                filename = os.path.basename(dlc_href)
                download_url = urllib.parse.urljoin(url, dlc_href)
                if not dlc.has_attr('title'):
                    file_format, file_size, last_update = 'Poly', None, None
                else:
                    if len(dlc.contents) < 3:
                        file_format, file_size = 'Txt', None
                    else:
                        file_format, file_size, _ = dlc.contents  # File type and size
                        file_format, file_size = file_format.strip(), file_size.text
                    last_update = pd.to_datetime(dlc.get('title'))  # Date and time
                parsed_dat = [filename, download_url, file_format, file_size, last_update]
                return parsed_dat

            subregion_download_catalogue = pd.DataFrame(parse_dlc(x) for x in download_links_class)
            subregion_download_catalogue.columns = ['Filename', 'URL', 'DataType', 'Size', 'LastUpdate']

            print("Done. ") if verbose else ""

        except Exception as e:
            subregion_download_catalogue = None
            print("Failed. {}".format(subregion_name_, e)) if verbose else ""

        return subregion_download_catalogue

def _get_download_index(update=False, confirmation_required=True, verbose=False):

    path_to_pickle = cd_dat(BBBikeDownloadDictName.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_pickle) and not update:
        download_dictionary = load_pickle(path_to_pickle)

    else:
        if confirmed("To collect {} from BBBike's free download server?".format(
                BBBikeDownloadDictName), confirmation_required=confirmation_required):

            try:
                bbbike_subregion_names =_get_subregion_catalogue(verbose=verbose).Name.to_list()

                if verbose:
                    print("Collecting {} ... ".format(BBBikeDownloadDictName))

                download_catalogue = [
                    _get_subregion_download_catalogue(subregion_name,confirmation_required=False,verbose=verbose)
                    for subregion_name in bbbike_subregion_names]

                sr_name = bbbike_subregion_names[0]
                sr_download_catalogue = download_catalogue[0]

                # Available file formats
                file_fmt = [re.sub('{}|CHECKSUM'.format(sr_name), '', f) for f in sr_download_catalogue.Filename]

                # Available data types
                data_typ = sr_download_catalogue.DataType.tolist()

                download_dictionary = {'FileFormat': [x.replace(".osm", "", 1) for x in file_fmt[:-2]],
                                       'DataType': data_typ[:-2],
                                       'Catalogue':dict(zip(bbbike_subregion_names, download_catalogue))}

                print("Finished. ") if verbose else ""

                save_pickle(download_dictionary, path_to_pickle, verbose=verbose)

            except Exception as e:
                print("Failed. {}".format(e))
                download_dictionary = None

        else:
            if verbose:
                print("No data of \"{}\" is available.".format(BBBikeDownloadDictName))
            download_dictionary = None

    return download_dictionary

def get_osm_file_formats():

    osm_file_formats =_get_download_index()['FileFormat']

    return osm_file_formats

def _validate_input_file_format(osm_file_format):

    assert isinstance(osm_file_format, str)
    bbbike_osm_file_formats =get_osm_file_formats()

    try:
        osm_file_format_ = find_similar_str(osm_file_format, bbbike_osm_file_formats)

        if osm_file_format_:
            return osm_file_format_

        else:
            print("The input file format must be one of the following:"
                  " \n  \"{}\".".format("\",\n  \"".join(bbbike_osm_file_formats)))

    except Exception as e:
        print(e)

def get_subregion_download_url(subregion_name, osm_file_format):

    subregion_name_ = _validate_input_subregion_name(subregion_name)
    osm_file_format_ = ".osm" + _validate_input_file_format(osm_file_format)

    bbbike_download_dictionary = _get_download_index()['Catalogue']
    sub_download_catalogue = bbbike_download_dictionary[subregion_name_]

    tmp = subregion_name_ + osm_file_format_
    url = sub_download_catalogue[sub_download_catalogue.Filename == tmp].URL.iloc[0]

    return subregion_name_, url

def _get_valid_download_info(subregion_name, osm_file_format, download_dir=None):

    subregion_name_, download_url = get_subregion_download_url(subregion_name, osm_file_format)
    osm_filename = os.path.basename(download_url)

    if download_dir:
        path_to_file = cd(validate_input_data_dir(download_dir), osm_filename,
                          mkdir=True)
    else:
        # default directory of package data
        path_to_file = cd_dat_bbbike(subregion_name_, osm_filename, mkdir=True)

    return subregion_name_, osm_filename, download_url, path_to_file

def download_osm_data_from_bbbike(subregion_names, osm_file_format='.osm.pbf', download_dir='osmfile',update=False,
                      confirmation_required=False, interval_sec=10,verbose=False,ret_download_path=False):
    if isinstance(subregion_names,str):
        subregion_names_=[subregion_names]
    elif isinstance(subregion_names,list):
        subregion_names_=subregion_names
    else:
        subregion_names_=[]
        print(" 'subregion_names' should be valued by str or list ")

    subregion_names_ = [_validate_input_subregion_name(x) for x in subregion_names_]

    osm_file_format_ = _validate_input_file_format(osm_file_format)

    download_path = []

    if confirmed("Confirmed to download {} data of the following geographic region(s):"
                 "\n\t{}\n?".format(osm_file_format_, "\n\t".join(subregion_names_)),
                 confirmation_required=confirmation_required):

        for sub_reg_name in subregion_names_:
            subregion_name_, osm_filename, download_url, path_to_file = \
                _get_valid_download_info(sub_reg_name, osm_file_format_,download_dir)

            if os.path.isfile(path_to_file) and not update:
                if verbose:
                    print(
                        "The {} data of {} is already available at \"\\{}\".".format(
                            osm_file_format_, subregion_name_,
                            os.path.relpath(path_to_file)))

                download_path.append(path_to_file)

            else:
                try:
                    if verbose:
                        print("{} \"{}\" to \"\\{}\" ... ".format(
                            "Updating" if os.path.isfile(path_to_file)
                            else "Downloading",
                            osm_filename,
                            os.path.relpath(os.path.dirname(path_to_file))))

                    download_osmfile_from_url(url=download_url,osm_filename=path_to_file)

                    print("Done. ") if verbose else ""

                    download_path.append(path_to_file)

                    if os.path.getsize(path_to_file) / (1024 ** 2) <= 5:
                        time.sleep(interval_sec)

                except Exception as e:
                    print("Failed. {}.".format(e))

        if ret_download_path:

            return download_path

def download_subregion_data_from_bbbike(subregion_name, download_dir='osmfile', update=False,confirmation_required=False,
                           verbose=True,ret_download_path=False):


    subregion_name_ = _validate_input_subregion_name(subregion_name)
    bbbike_download_dictionary =_get_download_index()['Catalogue']
    sub_download_catalogue = bbbike_download_dictionary[subregion_name_]

    data_dir = validate_input_data_dir(download_dir) if download_dir \
        else cd_dat_bbbike(subregion_name_, mkdir=True)

    if confirmed("Confirmed to download all available BBBike OSM data of {}?".format(
            subregion_name_), confirmation_required=confirmation_required):

        if verbose:
            if confirmation_required:
                print("Downloading in progress ... ")
            else:
                print("Downloading all available BBBike OSM data of {} ... ".format(
                    subregion_name_))

        download_paths = []

        for download_url, osm_filename in zip(sub_download_catalogue.URL,
                                              sub_download_catalogue.Filename):
            try:
                path_to_file = os.path.join(
                    data_dir, "" if not download_dir
                    else subregion_name_, osm_filename)

                if os.path.isfile(path_to_file) and not update:
                    if verbose:
                        print("\t\"{}\" is already available.".format(
                            os.path.basename(path_to_file)))

                else:
                    print("\t{} ... ".format(osm_filename)) if verbose else ""

                    download_osmfile_from_url(url=download_url, osm_filename=path_to_file)

                download_paths.append(path_to_file)

            except Exception as e:
                print("Failed. {}.".format(e))

        if verbose and download_paths:
            print("Done. Check out the downloaded OSM data at \"\\{}\".".format(
                os.path.relpath(os.path.dirname(download_paths[0]))))

        if ret_download_path:
            return download_paths

def update_bbbike_database():
    _get_list_of_subregion_names(update=True)


import bs4
import pandas as pd
from .utils import *
from .settings import *

def _get_raw_directory_index(url, verbose=False):
    try:
        import humanfriendly

        raw_directory_index = pd.read_html(url, match='file', header=0, parse_dates=['date'])
        raw_directory_index = pd.concat(raw_directory_index, ignore_index=True)
        raw_directory_index.columns = [c.title() for c in raw_directory_index.columns]

        # Clean the DataFrame
        raw_directory_index.Size = raw_directory_index.Size.apply(humanfriendly.format_size)
        raw_directory_index.sort_values('Date', ascending=False, inplace=True)
        raw_directory_index.index = range(len(raw_directory_index))

        raw_directory_index['FileURL'] = raw_directory_index.File.map(
            lambda x: urllib.parse.urljoin(url, x))

    except (urllib.error.HTTPError, TypeError, ValueError):
        if len(urllib.parse.urlparse(url).path) <= 1 and verbose:
            print("The web page does not have a raw directory index.")
        raw_directory_index = None

    return raw_directory_index

def _get_subregion_table(url, verbose=False):
        try:
            subregion_table = pd.read_html(
                url, match=re.compile(r'(Special )?Sub[ \-]Regions?'), encoding='UTF-8')
            subregion_table = pd.concat(subregion_table, axis=0, ignore_index=True)

            # Specify column names
            file_formats = GeofabrikValidFileFormats
            column_names = ['Subregion'] + file_formats
            column_names.insert(2, '.osm.pbf.Size')

            # Add column/names
            if len(subregion_table.columns) == 4:
                subregion_table.insert(2, '.osm.pbf.Size', np.nan)
            subregion_table.columns = column_names

            subregion_table.replace(
                {'.osm.pbf.Size': {re.compile('[()]'): '', re.compile('\xa0'): ' '}},
                inplace=True)

            # Get the URLs
            source = requests.get(url, headers=fake_requests_headers())
            soup = bs4.BeautifulSoup(source.content, 'lxml')
            source.close()

            for file_type in file_formats:
                text = '[{}]'.format(file_type)
                urls = [urllib.parse.urljoin(url, link['href']) for link in
                        soup.find_all(name='a', href=True, text=text)]
                subregion_table.loc[
                    subregion_table[file_type].notnull(), file_type] = urls

            try:
                subregion_urls = [
                    urllib.parse.urljoin(url, soup.find('a', text=text).get('href'))
                    for text in subregion_table.Subregion]
            except (AttributeError, TypeError):
                subregion_urls = [kml['onmouseover']
                                  for kml in soup.find_all('tr', onmouseover=True)]
                subregion_urls = [
                    s[s.find('(') + 1:s.find(')')][1:-1].replace('kml', 'html')
                    for s in subregion_urls]
                subregion_urls = [urllib.parse.urljoin(url, sub_url)
                                  for sub_url in subregion_urls]
            subregion_table['SubregionURL'] = subregion_urls

            column_names = list(subregion_table.columns)
            column_names.insert(1, column_names.pop(len(column_names) - 1))
            subregion_table = subregion_table[column_names]

            subregion_table['.osm.pbf.Size'] = \
                subregion_table['.osm.pbf.Size'].str.replace('(', '').str.replace(')', '')

            subregion_table = subregion_table.where(pd.notnull(subregion_table), None)

        except (ValueError, TypeError, ConnectionRefusedError, ConnectionError):
            # No more data available for subregions within the region
            if verbose:
                print("Checked out \"{}\".".format(
                    url.split('/')[-1].split('.')[0].title()))
            subregion_table = None

        return subregion_table

def _get_download_index(update=False, confirmation_required=True, verbose=False):
    path_to_download_index = cd_dat(GeofabrikDownloadIndexName.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_download_index) and not update:
        download_index = load_pickle(path_to_download_index)

    else:
        if confirmed("To get {}?".format(GeofabrikDownloadIndexName),
                     confirmation_required=confirmation_required):

            if verbose == 2:
                print("Collecting {}".format(GeofabrikDownloadIndexName), end=" ... ")
            try:
                import geopandas as gpd

                download_index_ = gpd.read_file(GeofabrikDownloadIndexURL)

                # Note that '<br />' exists in all the names of Poland' subregions
                download_index_.name = download_index_.name.str.replace('<br />', ' ')

                urls = download_index_.urls.map(
                    lambda x: pd.DataFrame.from_dict(x, 'index').T)
                urls_ = pd.concat(urls.values, ignore_index=True)
                download_index = pd.concat([download_index_, urls_], axis=1)

                print("Done. ") if verbose == 2 else ""

                save_pickle(download_index, path_to_download_index, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                download_index = None

        else:
            download_index = None
            if verbose:
                print("No data of {} is available.".format(GeofabrikDownloadIndexName))

    return download_index

def _get_continents_subregion_tables(update=False, confirmation_required=True, verbose=False):
    path_to_pickle =cd_dat(GeofabrikContinentSubregionTableName.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_pickle) and not update:
        subregion_tables = load_pickle(path_to_pickle)

    else:
        if confirmed("To collect information of {}?".format(
                GeofabrikContinentSubregionTableName),
                confirmation_required=confirmation_required):

            if verbose == 2:
                print("Collecting a table of {}".format(
                    GeofabrikContinentSubregionTableName), end=" ... ")

            try:
                # Scan the homepage to collect info of regions for each continent
                user_agent = fake_requests_headers()
                source = requests.get(GeofabrikURL, headers={"User-Agent": user_agent})
                soup = bs4.BeautifulSoup(source.text, 'lxml').find_all(
                    'td', {'class': 'subregion'})
                source.close()
                continent_names = [td.a.text for td in soup]
                continent_links = [urllib.parse.urljoin(GeofabrikURL, td.a['href'])
                                   for td in soup]
                subregion_tables = dict(
                    zip(continent_names,
                        [_get_subregion_table(url, verbose)
                         for url in continent_links]))

                print("Done. ") if verbose == 2 else ""

                save_pickle(subregion_tables, path_to_pickle, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                subregion_tables = None

        else:
            subregion_tables = None
            if verbose:
                print(f"No data of {GeofabrikContinentSubregionTableName} is available.")

    return subregion_tables

def _get_region_subregion_tier(update=False, confirmation_required=True,verbose=False):

    path_to_file = cd_dat(GeofabrikRegionSubregionTier.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_file) and not update:
        region_subregion_tier, non_subregions = load_pickle(path_to_file, verbose=verbose)
    else:
        def compile_region_subregion_tier(sub_reg_tbls):

            having_subregions = sub_reg_tbls.copy()
            region_subregion_tiers = having_subregions.copy()

            non_subregions_list = []
            for k, v in sub_reg_tbls.items():
                if v is not None and isinstance(v, pd.DataFrame):
                    region_subregion_tiers =update_nested_dict(sub_reg_tbls, {k: set(v.Subregion)})
                else:
                    non_subregions_list.append(k)

            for x in non_subregions_list:
                having_subregions.pop(x)

            having_subregions_temp = copy.deepcopy(having_subregions)

            while having_subregions_temp:
                for region_name, subregion_table in having_subregions.items():
                    subregion_names = subregion_table.Subregion
                    subregion_links = subregion_table.SubregionURL
                    sub_subregion_tables = dict(
                        zip(subregion_names,
                            [_get_subregion_table(link)
                             for link in subregion_links]))

                    subregion_index, without_subregion_ =compile_region_subregion_tier(sub_subregion_tables)
                    non_subregions_list += without_subregion_

                    region_subregion_tiers.update({region_name: subregion_index})

                    having_subregions_temp.pop(region_name)

            # Russian Federation in both pages of Asia and Europe,
            # so there are duplicates in non_subregions_list
            import more_itertools

            non_subregions_list = list(more_itertools.unique_everseen(non_subregions_list))

            return region_subregion_tiers, non_subregions_list

        if confirmed("To compile {}? (Note this may take up to a few minutes.)".format(GeofabrikRegionSubregionTier),
                     confirmation_required=confirmation_required):

            if verbose == 2:
                print("Compiling {} ... ".format(GeofabrikRegionSubregionTier), end="")

            # Scan the download pages to collect a catalogue of region-subregion tier
            try:
                subregion_tables =_get_continents_subregion_tables(update=update)
                region_subregion_tier, non_subregions = compile_region_subregion_tier(subregion_tables)

                print("Done. ") if verbose == 2 else ""

                save_pickle((region_subregion_tier, non_subregions), path_to_file,
                            verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                region_subregion_tier, non_subregions = None, None

        else:
            region_subregion_tier, non_subregions = None, None
            if verbose:
                print("No data of {} is available.".format(GeofabrikRegionSubregionTier))

    return region_subregion_tier, non_subregions

def _get_download_catalogue(update=False, confirmation_required=True, verbose=False):

    path_to_downloads_catalogue = cd_dat(GeofabrikDownloadsCatalogue.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_downloads_catalogue) and not update:
        subregion_downloads_catalogue = load_pickle(path_to_downloads_catalogue)

    else:
        if confirmed("To collect {}? (Note that it may take a few minutes.)".format(GeofabrikDownloadsCatalogue),
                     confirmation_required=confirmation_required):

            if verbose == 2:
                print("Collecting {}".format(GeofabrikDownloadsCatalogue), end=" ... ")
            try:
                user_agent = fake_requests_headers()
                source = requests.get(GeofabrikURL, headers={"User-Agent": user_agent})
                soup = bs4.BeautifulSoup(source.text, 'lxml')
                source.close()

                subregion_href = soup.find_all('td', {'class': 'subregion'})
                avail_subregion_urls = (urllib.parse.urljoin(GeofabrikURL, td.a['href'])
                                        for td in subregion_href)
                avail_subregion_url_tables_0 = (_get_subregion_table(sub_url, verbose)
                                                for sub_url in avail_subregion_urls)
                avail_subregion_url_tables = [tbl for tbl in avail_subregion_url_tables_0 if tbl is not None]

                subregion_url_tables = list(avail_subregion_url_tables)

                while subregion_url_tables:

                    subregion_url_tables_ = []

                    for subregion_url_table in subregion_url_tables:
                        subregion_urls = list(subregion_url_table.SubregionURL)
                        subregion_url_tables_0 = [
                            _get_subregion_table(sr_url, verbose)
                            for sr_url in subregion_urls]
                        subregion_url_tables_ += [
                            tbl for tbl in subregion_url_tables_0 if tbl is not None]

                        avail_subregion_url_tables += subregion_url_tables_

                    subregion_url_tables = list(subregion_url_tables_)

                # All available URLs for downloading
                home_subregion_url_table = _get_subregion_table(GeofabrikURL)
                avail_subregion_url_tables.append(home_subregion_url_table)
                subregion_downloads_catalogue = pd.concat(avail_subregion_url_tables,
                                                          ignore_index=True)
                subregion_downloads_catalogue.drop_duplicates(inplace=True)

                duplicated = subregion_downloads_catalogue[subregion_downloads_catalogue.Subregion.duplicated(keep=False)]
                if not duplicated.empty:
                    import humanfriendly

                    for i in range(0, 2, len(duplicated)):
                        temp = duplicated.iloc[i:i + 2]
                        size = temp['.osm.pbf.Size'].map(lambda x: humanfriendly.parse_size(
                            x.strip('(').strip(')').replace('\xa0', ' ')))
                        idx = size[size == size.min()].index
                        subregion_downloads_catalogue.drop(idx, inplace=True)
                    subregion_downloads_catalogue.index = range(len(subregion_downloads_catalogue))

                # Save subregion_index_downloads to local disk
                save_pickle(subregion_downloads_catalogue,
                            path_to_downloads_catalogue, verbose=verbose)

            except Exception as e:
                print("Failed. {}.".format(e))
                subregion_downloads_catalogue = None

        else:
            subregion_downloads_catalogue = None
            if verbose:
                print("No data of {} is available.".format(GeofabrikDownloadsCatalogue))

    return subregion_downloads_catalogue

def _get_list_of_subregion_names(update=False, confirmation_required=True,verbose=False):

    path_to_name_list = cd_dat(GeofabrikSubregionNameList.replace(" ", "-") + ".pickle")

    if os.path.isfile(path_to_name_list) and not update:
        subregion_name_list = load_pickle(path_to_name_list)

    else:
        if confirmed("To get {}?".format(GeofabrikSubregionNameList),confirmation_required=confirmation_required):

            downloads_catalogue = _get_download_catalogue(update=update, confirmation_required=False)

            subregion_name_list = downloads_catalogue.Subregion.to_list()

            save_pickle(subregion_name_list, path_to_name_list, verbose=verbose)
        else:
            subregion_name_list = []
            if verbose:
                print("No data of {} is available.".format(GeofabrikSubregionNameList))
    return subregion_name_list

def _validate_input_subregion_name(subregion_name):

    assert isinstance(subregion_name, str)
    # Get a list of available
    subregion_names = _get_list_of_subregion_names()

    if os.path.isdir(os.path.dirname(subregion_name)) or urllib.parse.urlparse(subregion_name).path:
        subregion_name_ = find_similar_str(os.path.basename(subregion_name),
                                           subregion_names)
    else:
        subregion_name_ = find_similar_str(subregion_name, subregion_names)

    if not subregion_name_:
        raise ValueError(
            "The input subregion name is not identified.\n"
            "Check if the required subregion exists in the catalogue and retry.")

    return subregion_name_

def _validate_input_file_format(osm_file_format):

    osm_file_format_ = find_similar_str(osm_file_format,GeofabrikValidFileFormats)

    assert osm_file_format_ in GeofabrikValidFileFormats,"The input file format must be one from {}.".format(
        GeofabrikValidFileFormats
    )

    return osm_file_format_

def _get_subregion_download_url(subregion_name, osm_file_format, update=False,verbose=False):

    # Get an index of download URLs
    subregion_downloads_index =_get_download_catalogue(update=update, verbose=verbose)
    subregion_downloads_index.set_index('Subregion', inplace=True)

    subregion_name_ = _validate_input_subregion_name(subregion_name)
    osm_file_format_ = _validate_input_file_format(osm_file_format)

    # Get the URL
    download_url = subregion_downloads_index.loc[subregion_name_, osm_file_format_]

    return subregion_name_, download_url

def _get_default_osm_filename(subregion_name, osm_file_format, update=False):

    subregion_name_ = _validate_input_subregion_name(subregion_name)
    osm_file_format_ =_validate_input_file_format(osm_file_format)

    _, download_url = _get_subregion_download_url(
        subregion_name_, osm_file_format_, update=update)

    if download_url is None:
        print("No {} data is available to download for {}.".format(
            osm_file_format_, subregion_name_))

    else:
        subregion_filename = os.path.split(download_url)[-1]
        return subregion_filename

def _get_default_path_to_osm_file(subregion_name, osm_file_format, mkdir=False,update=False, verbose=False):

    subregion_name_ = _validate_input_subregion_name(subregion_name)
    osm_file_format_ = _validate_input_file_format(osm_file_format)

    subregion_name_, download_url = _get_subregion_download_url(
        subregion_name_, osm_file_format_, update=update)

    if download_url is None:
        if verbose:
            print("{} data is not available for {}".format(
                osm_file_format_, subregion_name_))

        default_filename, default_file_path = None, None

    else:
        parsed_path = urllib.parse.urlparse(download_url).path.lstrip('/').split('/')

        if len(parsed_path) == 1:
            parsed_path = [subregion_name_] + parsed_path

        subregion_names = _get_list_of_subregion_names()
        directory = cd_dat_geofabrik(*[find_similar_str(x, subregion_names) if x != 'us' else 'United States'
                                       for x in parsed_path[0:-1]],
                                     mkdir=mkdir)

        default_filename = parsed_path[-1]
        default_file_path = os.path.join(directory, default_filename)

    return default_filename, default_file_path

def _search_for_subregions( *subregion_name, deep=False):

    region_subregion_tier, non_subregions_list = _get_region_subregion_tier()

    if not subregion_name:
        subregion_names = non_subregions_list

    else:

        def find_subregions(reg_name, reg_sub_idx):

            for k, v in reg_sub_idx.items():
                if reg_name == k:
                    if isinstance(v, dict):
                        yield list(v.keys())
                    else:
                        yield [reg_name] if isinstance(reg_name, str) else reg_name
                elif isinstance(v, dict):
                    for sub in find_subregions(reg_name, v):
                        if isinstance(sub, dict):
                            yield list(sub.keys())
                        else:
                            yield [sub] if isinstance(sub, str) else sub

        res = []
        for region in subregion_name:
            res += list(find_subregions(_validate_input_subregion_name(region),region_subregion_tier))[0]

        if not deep:
            subregion_names = res
        else:
            check_list = [x for x in res if x not in non_subregions_list]
            if check_list:
                res_ = list(set(res) - set(check_list))
                res_ += _search_for_subregions(*check_list)
            else:
                res_ = res
            del non_subregions_list, region_subregion_tier, check_list

            subregion_names = list(dict.fromkeys(res_))

    return subregion_names

def _make_sub_download_dir(subregion_name, osm_file_format, download_dir=None, mkdir=False):

    subregion_name_ =_validate_input_subregion_name(subregion_name)
    osm_file_format_ =_validate_input_file_format(osm_file_format)

    default_filename, default_file_path = _get_default_path_to_osm_file(
        subregion_name_, osm_file_format_)

    if not default_filename:
        default_sub_dir = re.sub( r"[. ]", "-", subregion_name_.lower() + osm_file_format_)
    else:
        default_sub_dir = re.sub(r"[. ]", "-", default_filename).lower()

    if not download_dir:
        default_download_dir = cd_dat_geofabrik(os.path.dirname(default_file_path),default_sub_dir, mkdir=mkdir)

    else:
        default_download_dir = cd(validate_input_data_dir(download_dir),default_sub_dir, mkdir=mkdir)

    return default_download_dir

def osm_file_exists(subregion_name, osm_file_format, data_dir=None,update=False,verbose=False, ret_file_path=False):
    subregion_name_ = _validate_input_subregion_name(subregion_name)
    osm_file_format_ =_validate_input_file_format(osm_file_format)

    default_filename, path_to_file = _get_default_path_to_osm_file(
        subregion_name_, osm_file_format_)

    if data_dir:
        path_to_file = cd(validate_input_data_dir(data_dir), default_filename)

    if os.path.isfile(path_to_file) and not update:
        if verbose == 2:
            print("\"{}\" of {} is available at \"{}\".".format(
                default_filename, subregion_name_,
                os.path.relpath(os.path.dirname(path_to_file))))

        if ret_file_path:
            return path_to_file
        else:
            return True
    else:
        return False

def download_osm_data_from_geofabrik(subregion_names, osm_file_format, download_dir='osmfile',update=False,
                      confirmation_required=False, deep_retry=False,interval_sec=10, verbose=True,
                      ret_download_path=False):

    subregion_names_ = [subregion_names] if isinstance(subregion_names, str) else subregion_names.copy()
    subregion_names_ = [_validate_input_subregion_name(x) for x in subregion_names_]

    osm_file_format_ = _validate_input_file_format(osm_file_format)

    if confirmed(
            "Confirmed to download {} data of the following geographic region(s):"
            "\n\t{}\n?".format(osm_file_format_, "\n\t".join(subregion_names_)),
            confirmation_required=confirmation_required):

        download_paths = []

        for sub_reg_name in subregion_names_:

            # Get download URL
            subregion_name_, download_url = _get_subregion_download_url(
                sub_reg_name, osm_file_format_)

            if download_url is None:

                if verbose:
                    print("The {} data is not found for \"{}\".".format(
                        osm_file_format_, subregion_name_))

                if confirmed("Try downloading the data of its subregions instead",
                             confirmation_required=confirmation_required):

                    sub_subregions = _search_for_subregions(
                        subregion_name_, deep=deep_retry)

                    if sub_subregions == [subregion_name_]:
                        print("No {} data is available "
                              "for this geographic region.".format(osm_file_format_))
                        break

                    else:
                        if not download_dir:
                            _, path_to_file_ = _get_default_path_to_osm_file(
                                subregion_name_, ".osm.pbf")
                            download_dir = os.path.dirname(path_to_file_)

                        download_dir_ = _make_sub_download_dir(
                            subregion_name_, osm_file_format_, download_dir)

                        download_osm_data_from_geofabrik(
                            sub_subregions, osm_file_format=osm_file_format_,
                            download_dir=download_dir_, update=update,
                            confirmation_required=False, verbose=verbose,
                            ret_download_path=ret_download_path)

            else:
                if not download_dir:
                    # Download the requested OSM file to default directory
                    osm_filename, path_to_file = _get_default_path_to_osm_file(
                        subregion_name_, osm_file_format_, mkdir=True)
                else:
                    download_dir_ = validate_input_data_dir(download_dir)
                    osm_filename = _get_default_osm_filename(
                        subregion_name_, osm_file_format=osm_file_format_)
                    path_to_file = os.path.join(download_dir_, osm_filename)

                download_paths.append(path_to_file)

                if os.path.isfile(path_to_file) and not update:
                    if verbose:
                        print("\"{}\" is already available at \"\\{}\".".format(
                            os.path.basename(path_to_file),
                            os.path.relpath(os.path.dirname(path_to_file))))

                else:
                    if verbose:
                        print("{} \"{}\" to \"\\{}\" ... ".format(
                            "Updating" if os.path.isfile(path_to_file)
                            else "Downloading",
                            osm_filename,
                            os.path.relpath(os.path.dirname(path_to_file))))

                    try:
                        download_osmfile_from_url(url=download_url, osm_filename=path_to_file)
                        print("Done. ") if verbose else ""
                    except Exception as e:
                        print("Failed. {}.".format(e))

            if len(subregion_names_)>1 and interval_sec:
                time.sleep(interval_sec)

        if ret_download_path:

            return download_paths

def download_subregion_data_from_geofabrik(subregion_names, osm_file_format, download_dir=None, update=False,verbose=True,
                            ret_download_path=False):

    subregion_names_ = [subregion_names] if isinstance(subregion_names, str) else subregion_names.copy()
    subregion_names_ = [_validate_input_subregion_name(x) for x in subregion_names_]
    subregion_names_ = _search_for_subregions(*subregion_names_)

    subregion_name_list = subregion_names_.copy()

    osm_file_format_ = _validate_input_file_format(osm_file_format)

    for subregion_name in subregion_names_:
        if osm_file_exists(subregion_name, osm_file_format_, download_dir, update):
            subregion_name_list.remove(subregion_name)

    confirmation_required_ = False if not subregion_name_list else True

    if confirmed(
            "Confirmed to download {} data of the following geographic region(s): "
            "\n\t{}\n?".format(osm_file_format_, "\n\t".join(subregion_name_list)),
            confirmation_required=confirmation_required_):

        download_paths = download_osm_data_from_geofabrik(
            subregion_names_, osm_file_format=osm_file_format_,
            download_dir=download_dir, update=update, confirmation_required=False,
            verbose=verbose, ret_download_path=ret_download_path)

        if ret_download_path:
            return download_paths

def update_geofabrik_database():
    _get_list_of_subregion_names(update=True)


# python setup.py sdist
# python setup.py bdist_wheel
# twine upload dist/*0.2.1*

import setuptools

setuptools.setup(
    name='osm2rail',
    version='0.0.3',
    author='Jiawei Lu, Qian Fu, Zanyang Cui, Dr.Junhua Chen',
    author_email='jiaweil9@asu.edu, q.fu@bham.ac.uk, zanyangcui@outlook.com, cjh@bjtu.edu.cn',
    url='https://github.com/PariseC/osm2rail',
    description='An open-source education tool for constructing modeling datasets of railway transportation',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    license='Apache License 2.0',
    packages=['osm2rail'],
    include_package_data=True,
    python_requires=">=3.6.0",
    install_requires=[
        'bs4',
        'matplotlib<=3.3.0',
        'osmium',
        'fuzzywuzzy'
    ],
    classifiers=['License :: OSI Approved :: Apache Software License',
                 'Programming Language :: Python :: 3']
)



from .utils import *
from .web import *

def _get_download_url(box):
    URL = overpass_homepage()
    min_lat, max_lat, min_lon, max_lon = box
    url = URL + str(min_lon) + ',' + str(min_lat) + ',' + str(max_lon) + ',' + str(max_lat)
    return url

def _make_download_file(subarea_name, download_dir):
    path_to_file = os.path.join(download_dir, subarea_name) + '.osm'
    if os.path.isfile(path_to_file):
        print("\"{}\" is already available at \"\\{}\".".format(
            os.path.basename(path_to_file),
            os.path.relpath(os.path.dirname(path_to_file))))
        return True,path_to_file
    return False,path_to_file

def download_osm_data_from_overpass(subarea_names=None,bboxs=None,download_dir='osmfile',
                               interval_sec=10,ret_download_path=False):

    download_dir_ = validate_download_dir(download_dir)
    download_paths = []

    if subarea_names:
        if isinstance(subarea_names,str):
            subarea_names_=[subarea_names]
        elif isinstance(subarea_names,list):
            subarea_names_=subarea_names
        else:
            subarea_names_=[]
            print(" 'subarea_names' should be valued by str or list ")

        for subarea_name in subarea_names_:
            subarea_name_=subarea_name.replace(' ','_')
            is_downloaded,osm_filename = _make_download_file(subarea_name_, download_dir_)
            download_paths.append(osm_filename)
            if not is_downloaded:
                box = get_subregion_download_range(subarea_name)
                print("The box of {} is {} ".format(subarea_name_,box))
                print("Downloading \"{}.osm\" to \"\\{}\"...".format(subarea_name_,os.path.relpath(os.path.dirname(osm_filename))))
                if box:
                    url = _get_download_url(box)
                    try:
                        download_osmfile_from_url(url=url,osm_filename=osm_filename)
                        print('Done.')
                    except Exception as e:
                        print("Failed. {}.".format(e))
                    if interval_sec and len(subarea_names_)>1:
                        time.sleep(interval_sec)
    if bboxs:
        if isinstance(bboxs,tuple):
            bboxs_=[bboxs]
        elif isinstance(bboxs,list):
            bboxs_=bboxs
        else:
            bboxs_=[]
            print(" 'bboxs' should be valued by tuple or list ")

        for id,box in enumerate(bboxs_):
            subbox_name='map_'+str(id+1)
            is_downloaded,osm_filename = _make_download_file(subbox_name, download_dir_)
            download_paths.append(osm_filename)
            if not is_downloaded:
                print("Downloading \"{}.osm\" to \"\\{}\"...".format(subbox_name,os.path.relpath(os.path.dirname(osm_filename))))
                try:
                    url = _get_download_url(box)
                    download_osmfile_from_url(url=url,osm_filename=osm_filename)
                    print('Done.')
                except Exception as e:
                    print("Failed. {}.".format(e))
            if interval_sec and len(bboxs_)>1:
                time.sleep(interval_sec)
    if ret_download_path:
        return download_paths
